/*  gentrex - generic tree expressions; minimal test tree for the example_app
    Copyright (C) 2015  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page: http://repo.hu/projects/gentrex
    Contact the author (email and IRC): http://igor2.repo.hu/contact.html
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "test_tree.h"

/* example tree
  .name      .uid    .user
--------------------------
	A          0
		B        1
			E      4
			F      5
			G      6
		C        2
			H      7
				N    13
				O    14
			I      8
				P    15
				Q    16
				R    17
			J      9
				S    18
		D        3
			K      10
				T    19
				U    20
			L      11
				V    22
				W    23
					Y  25
					Z  26
			M      12
				X    24
*/

/***** tree *****/

char *tree_user[26];

int tree_child_tab[26] = {
	 /*A=*/T('B'), /*B=*/T('E'), /*C=*/T('H'), /*D=*/T('K'), /*E=*/-1
	,/*F=*/-1    , /*G=*/-1    , /*H=*/T('N'), /*I=*/T('P'), /*J=*/T('S')
	,/*K=*/T('T'), /*L=*/T('V'), /*M=*/T('X'), /*N=*/-1    , /*O=*/-1
	,/*P=*/-1    , /*Q=*/-1    , /*R=*/-1    , /*S=*/-1    , /*T=*/-1
	,/*U=*/-1    , /*V=*/-1    , /*W=*/T('Y'), /*X=*/-1    , /*Y=*/-1
	,/*Z=*/-1
};

int tree_next_tab[26] = {
	 /*A=*/-1    , /*B=*/T('C'), /*C=*/T('D'), /*D=*/-1    , /*E=*/T('F')
	,/*F=*/T('G'), /*G=*/-1    , /*H=*/T('I'), /*I=*/T('J'), /*J=*/-1
	,/*K=*/T('L'), /*L=*/T('M'), /*M=*/-1    , /*N=*/T('O'), /*O=*/-1
	,/*P=*/T('Q'), /*Q=*/T('R'), /*R=*/-1    , /*S=*/-1    , /*T=*/T('U')
	,/*U=*/-1    , /*V=*/T('W'), /*W=*/-1    , /*X=*/-1    , /*Y=*/T('Z')
	,/*Z=*/-1
};

int tree_parent_tab[26];

void tree_parent_init(int root)
{
	int child;
	for(child = tree_child_tab[root]; child != -1; child = tree_next_tab[child]) {
		tree_parent_tab[child] = root;
		tree_parent_init(child);
	}
}

void tree_init(void)
{
	tree_parent_init(0);
	tree_parent_tab[0] = -1;
}

#define get_tree_idx(idx, node) \
	(idx = (char **)node - (char **)tree_user, ((idx >= 0) && (idx < sizeof(tree_user)/sizeof(tree_user[0])))) \

void *tree_child(void *user_ctx, void *node)
{
	int idx;
	assert(get_tree_idx(idx, node));

	if (tree_child_tab[idx] == -1)
		return NULL;
	return (char **)tree_user + tree_child_tab[idx];
}

void *tree_parent(void *user_ctx, void *node)
{
	int idx;
	assert(get_tree_idx(idx, node));

	if (tree_parent_tab[idx] == -1)
		return NULL;
	return (char **)tree_user + tree_parent_tab[idx];

}

void *tree_next(void *user_ctx, void *node)
{
	int idx;
	assert(get_tree_idx(idx, node));

	if (tree_next_tab[idx] == -1)
		return NULL;
	return (char **)tree_user + tree_next_tab[idx];
}

void tree_print_node(const char *prefix, void *node)
{
	int idx;

	printf("%s", prefix);

	if (node == NULL) {
		printf("<FALSE>\n");
		return;
	}

	if (!get_tree_idx(idx, node)) {
		printf("<invalid>\n");
		return;
	}

	printf("%c (%d): '%s'\n", idx+'A', idx,  tree_user[idx]);
}

void tree_print_nodecomp(const char *prefix, gentrex_nodecomp_t ndc)
{
	printf("%s\n", prefix);
	tree_print_node(" -> ", ndc.straight);
	tree_print_node(" -> ", ndc.inverted);
}

/***** tag *****/
char tagbuf[16];
static const char *tag_get(void *node, const char *tagname)
{
	if (strcmp(tagname, "uid") == 0) {
		sprintf(tagbuf, "%d", (char **)node - (char **)tree_user);
		return tagbuf;
	}
	if (strcmp(tagname, "name") == 0) {
		tagbuf[0] = 'A' + ((char **)node - (char **)tree_user);
		tagbuf[1] = '\0';
		return tagbuf;
	}
	if (strcmp(tagname, "user") == 0)
		return *(char **)node;
	return NULL;
}

int tag_eq(void *user_ctx, void *node, void *op_cookie, const char *tagname, const char *str)
{
	const char *val = tag_get(node, tagname);
	if (val == NULL)
		return 0;
	return strcmp(val, str) == 0;
}

/*int tag_match(void *user_ctx, void *node, const char *tagname, const char *pattern)
{
}*/
